/***************************************************************************
 *   Copyright (C) 2004 by Petric Frank                                    *
 *   pfrank@gmx.de                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "cvs-command-checkin.h"
#include "cvs-session.h"
#include "cvs-connection.h"
#include "cvs-directory-entry.h"



cvs_command_checkin::cvs_command_checkin (const std::string &path, const std::string &fullpath,
                                          const std::string &comment, cvs_directory_entry &entry)
: cvs_command (), item_path (path), item_fullpath (fullpath), item_comment (comment), item_entry (entry)
{
  // strip version - if any
  std::string::size_type pos = item_path.find ("@@");
  if (pos != std::string::npos)
    item_path.erase (pos);

  // strip away leading slash - if any
  if ((item_path.length () > 0) && (item_path[0] == '/'))
    item_path.erase (0, 1);

  memset (&item_stat, 0, sizeof (item_stat));
}



bool cvs_command_checkin::run (cvs_session &session)
{
  // reset the log filename
  item_entry.set_logfilename ("");
  
  // directories can not be checked in to CVS
  if (item_entry.is_dir ())
    return false;

  // obtain file info
  ::lstat (item_fullpath.c_str (), &item_stat);
//  if (item_entry.valid ())
//  {
//    struct stat stbuf;
//    item_entry.get_attributes (&stbuf);
//    item_stat.st_mode = stbuf.st_mode;
//  }
    
  // open source file for read
  cache_file.open (item_fullpath.c_str ());
  if (!cache_file.is_open ())
    return false;

  // store responses from CVS server
  if (establish_logging (session))
    item_entry.set_logfilename (filename);

  // send CVS command
  if (!send_command (session))
    return false;

  // obtain response from CVS server
  if (!process_response (session))
    return false;

  // update version number
  item_entry.set_version (new_version);

  // file stats are no more valid
  item_entry.invalidate ();

  return true;
}



bool cvs_command_checkin::send_command (cvs_session &session)
{
  session.send () << "Argument -m" << std::endl;

  std::string comment = item_comment;
  std::string::size_type pos = comment.find ('\n');
  std::string remain;
  if (pos != std::string::npos)
  {
    remain = comment;
    remain.erase (0, pos + 1);
    comment.erase (pos);
  }
  session.send () << "Argument " << comment << std::endl;
  
  while (remain.length () != 0)
  {
    comment = remain;
    pos = comment.find ('\n');
    if (pos != std::string::npos)
    {
      remain.erase (0, pos + 1);
      comment.erase (pos);
    }
    else
      remain.clear ();
    
    session.send () << "Argumentx " << comment << std::endl;
  }

  std::string path = session.get_connection ().cvsroot () + "/" + item_path;
  pos = path.rfind ('/');
  if (pos != std::string::npos)
    path.erase (pos);

  session.send () << "Argument --" << std::endl
                  << "Directory ." << std::endl
		  << path << std::endl
		  << "Entry /" << item_entry.name () << "/" << item_entry.version () << "///" << std::endl
		  << "Modified " << item_entry.name () << std::endl;

  std::string attrstr;
  sys2cvs_attr (attrstr, item_stat.st_mode);

  session.send () << attrstr << std::endl
		  << item_stat.st_size << std::endl;

  // now transfer the file contents
  char block[4096];
  int count;

  cache_file.read (block, 4096);
  while ((count = cache_file.gcount ()) > 0)
  {
    session.send ().write (block, count);
    
    cache_file.read (block, 4096);
  }

  session.send () << "Argument " << item_entry.name () << std::endl
                  << "ci" << std::endl;

  return true;
}



bool cvs_command_checkin::process_line (cvs_session & session, const std::string & line)
{
  const char InitVersionMarker[]	= "M initial revision: ";
  const char NewVersionMarker[]		= "M new revision: ";
  const char OldVersionMarker[]		= "; previous revision: ";
  const int InitVersionMarkerSize	= sizeof (InitVersionMarker) - 1;
  const int NewVersionMarkerSize	= sizeof (NewVersionMarker) - 1;

  // Revision notification (updated file)
  if (std::string::traits_type::compare (line.data (), NewVersionMarker, NewVersionMarkerSize) == 0)
  {
    new_version = line;
    new_version.erase (0, NewVersionMarkerSize);
    
    std::string::size_type pos = new_version.find (OldVersionMarker);
    if (pos != std::string::npos)
      new_version.erase (pos);

    return true;
  }

  // Revision notification (new file)
  if (std::string::traits_type::compare (line.data (), InitVersionMarker, InitVersionMarkerSize) == 0)
  {
    new_version = line;
    new_version.erase (0, NewVersionMarkerSize);
  }

  return true;
}



void cvs_command_checkin::sys2cvs_attr (std::string & line, int attr)
{
  line = "u=";
  if (attr & S_IRUSR)
    line += "r";
  if (attr & S_IWUSR)
    line += "w";
  if (attr & S_IXUSR)
    line += "x";

  line += ",g=";
  if (attr & S_IRGRP)
    line += "r";
  if (attr & S_IWGRP)
    line += "w";
  if (attr & S_IXGRP)
    line += "x";

  line += ",o=";
  if (attr & S_IROTH)
    line += "r";
  if (attr & S_IWOTH)
    line += "w";
  if (attr & S_IXOTH)
    line += "x";
}
