/***************************************************************************
 *   Copyright (C) 2004 by Petric Frank                                    *
 *   pfrank@gmx.de                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "cvs-command-checkout.h"
#include "cvs-session.h"
#include "cvs-connection.h"
#include "cvs-directory-entry.h"



cvs_command_checkout::cvs_command_checkout (const std::string &path, const std::string &dest,
                                            cvs_directory_entry &entry)
: cvs_command (), item_path (path), item_destination (dest), item_entry (entry), cache_file ()
{
  std::string::size_type pos = item_path.find ("@@");

  if (pos != std::string::npos)
    item_path.erase (pos);	// strip away version - if any

  // strip away leading slash - if any
  if ((item_path.length () > 0) && (item_path[0] == '/'))
    item_path.erase (0, 1);

  memset (&item_stat, 0, sizeof (item_stat));
}



bool cvs_command_checkout::run (cvs_session &session)
{
  // reset the log filename
  item_entry.set_logfilename ("");
  
  // directories can not be checked out by CVS
  if (item_entry.is_dir ())
    return false;

  // allocate the directory leading to the file
  std::string dir = item_destination;
  std::string::size_type pos = dir.rfind ('/');
  
  if (pos != std::string::npos)
    dir.erase (pos);
    
  if (!create_directory (dir))
    return false;

  // open destination file for write
  cache_file.open (item_destination.c_str ());
  if (!cache_file.is_open ())
    return false;

  // store responses from CVS server
  if (establish_logging (session))
    item_entry.set_logfilename (filename);

  // send CVS command
  if (!send_command (session))
    return false;

  // obtain response from CVS server
  if (!process_response (session))
    return false;

  item_entry.set_attributes (item_stat);

  return true;
}



bool cvs_command_checkout::send_command (cvs_session &session)
{
  session.send () << "Argument -N" << std::endl
                  << "Argument -P" << std::endl
		  << "Argument -r" << std::endl
		  << "Argument " << item_entry.version () << std::endl
		  << "Argument " << item_path << std::endl
		  << "Directory ." << std::endl
		  << session.get_connection ().cvsroot () << std::endl
		  << "co" << std::endl;

  return true;
}



bool cvs_command_checkout::process_line (cvs_session & session, const std::string & line)
{
  // file date ?
  if (line.find ("Mod-time ") == 0)
  {
    struct tm tm;
    int filetime = 0;

    memset (&tm, 0, sizeof (tm));
    if (strptime (&(line.c_str ())[9], "%d %b %Y %T ", &tm) != 0)
      filetime = mktime (&tm);

    item_stat.st_atime = filetime;
    item_stat.st_ctime = filetime;
    item_stat.st_mtime = filetime;

    return true;
  }

  // file mode ?
  if ((line.length () > 2) && (line[0] == 'u') && (line[1] == '='))
  {
    item_stat.st_mode = cvs2sys_attr (line) | S_IFREG;	// always regular files

    return true;
  }

  // file contents ?
  if (isdigit (line[0]) != 0)
  {
    item_stat.st_size = strtoul (line.c_str (), 0, 0);

    // now load the file contents
    const int bufsize = 4096;		// read the file in 4k blocks
    char readbuf[bufsize];
    loff_t pos = 0;
    loff_t size = item_stat.st_size;	// expected size of file

    while (pos < size)
    {
      int chunk = bufsize;
      if ((pos + chunk) > size)
        chunk = size - pos;

      int bytes_read = session.read_raw (readbuf, chunk);
      if (bytes_read == -1) // timeout
        return false;

//      if (bytes_read == 0)	// no more data
//        break;

      cache_file.write (readbuf, bytes_read);

      pos += bytes_read;
    }
  }

  return true;
}



int cvs_command_checkout::cvs2sys_attr (const std::string & line)
{
  const int bits[9] =  {S_IRUSR, S_IWUSR, S_IXUSR,
			S_IRGRP, S_IWGRP, S_IXGRP,
			S_IROTH, S_IWOTH, S_IXOTH};
  std::string::const_iterator iter = line.begin ();
  int stage = 0;
  int attr = 0;
  bool doStage = true;

  do
  {
    if (doStage)
    {
      switch (*iter)
      {
        case 'u':	stage = 0;		break;
        case 'g':	stage = 3;		break;
        case 'o':	stage = 6;		break;
        case '=':	doStage = false;	break;
        default:	return attr;
      }
    }
    else
    {
      switch (*iter)
      {
        case 'r':	attr |= bits[stage];	break;
        case 'w':	attr |= bits[stage+1];	break;
        case 'x':	attr |= bits[stage+2];	break;
	case ',':	doStage = true;		break;
	default:	return attr;
      }
    }

    ++iter;
  } while (iter != line.end ());

  return attr;
}



bool cvs_command_checkout::create_directory (const std::string &dir)
{
  std::string path = dir;
  std::string::size_type pos;
  std::string dirpart;
  int dirmode = S_IFDIR | S_IRUSR | S_IWUSR | S_IXUSR;

  pos = 0;
  do
  {
    pos = path.find ('/', pos + 1);

    dirpart = path;
    if (pos != std::string::npos)
      dirpart.erase (pos);

    struct stat info;

    if (::lstat (dirpart.c_str (), &info) != 0)
    {
      int result = ::mkdir (dirpart.c_str (), dirmode);
      if (result != 0)
        return false;
    }
    else
      if (!S_ISDIR (info.st_mode))
        return false;

  } while (pos != std::string::npos);

  return true;
}
