/***************************************************************************
 *   Copyright (C) 2004 by Petric Frank                                    *
 *   pfrank@gmx.de                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "cvs-directory-entry.h"
#include <stdio.h>
#include <string.h>
#include "system-log.h"



cvs_directory_entry::cvs_directory_entry (const std::string & name, const std::string &version, bool isdir)
: data (new entry_data (name, version, isdir))
{
}



cvs_directory_entry::cvs_directory_entry (const cvs_directory_entry &master)
: data (master.data)
{
  ++(data->refcount);
}



cvs_directory_entry::~cvs_directory_entry ()
{
  --(data->refcount);
  if (data->refcount == 0)
  {
    delete data;
    data = 0;
  }
}



cvs_directory_entry &cvs_directory_entry::operator = (const cvs_directory_entry &rhs)
{
  if (&rhs != this)
  {
    // does it point to the same data block - do nothing
    if (rhs.data != data)
    {
      entry_data *temp = data;
      
      data = rhs.data;
      ++(data->refcount);

      --(temp->refcount);
      if (temp->refcount == 0)
        delete temp;

    }
  }

  return *this;
}



std::ostream & cvs_directory_entry::operator << (std::ostream &os)
{
  // do not stream version 0 elements
  if (data->cvs_fileversion == "0")
    return os;

  os << data->cvs_filename << std::ends
     << data->cvs_fileversion << std::ends
     << data->is_directory
     << data->valid_attributes
     << data->log_filename << std::ends
     << data->cvs_filestats.st_mode << " "
     << data->cvs_filestats.st_atime << " "
     << data->contents.size () << std::endl;

  directory_list::iterator iter;
  
  for (iter = data->contents.begin (); iter != data->contents.end (); ++iter)
    os << *iter;

  return os;
}



std::istream & cvs_directory_entry::operator >> (std::istream &is)
{
  directory_list::size_type count;
  char dummy;

  if (!is.good ())
  {
    system_log *log = system_log::instance ();

    log->error << "Loading tree aborted - input stream failed" << std::endl;

    return is;
  }

  read_string (data->cvs_filename, is);
  read_string (data->cvs_fileversion, is);
  read_bool (data->is_directory, is);
  read_bool (data->valid_attributes, is);
  read_string (data->log_filename, is);
  is >> data->cvs_filestats.st_mode;
  is.get (dummy);	// drop delimiting blanc
  is >> data->cvs_filestats.st_atime;
  is.get (dummy);	// drop delimiting blanc
  is >> count;
  is.get (dummy);	// drop newline

  data->cvs_filestats.st_ctime = data->cvs_filestats.st_atime;
  data->cvs_filestats.st_mtime = data->cvs_filestats.st_atime;

  // read the subitems
  for (directory_list::size_type loop = 0; loop < count; ++loop)
  {
    cvs_directory_entry entry ("", "", false);

    is >> entry;

    add_entry (entry);
  }

  return is;
}



std::ostream & operator << (std::ostream &os, cvs_directory_entry &entry)
{
  return entry.operator << (os);
}



std::istream & operator >> (std::istream &is, cvs_directory_entry &entry)
{
  return entry.operator >> (is);
}



std::istream & cvs_directory_entry::read_string (std::string &item, std::istream &is)
{
  char c;

  item.clear ();

  is.read (&c, 1);
  while (is.good () && (c != '\0'))
  {
    item += c;
    is.read (&c, 1);
  }

  return is;
}



std::istream & cvs_directory_entry::read_bool (bool &item, std::istream &is)
{
  char c;

  is.read (&c, 1);
  
  item = c == '1';

  return is;
}



void cvs_directory_entry::set_attributes (const struct stat &stbuf)
{
  memcpy (&data->cvs_filestats, &stbuf, sizeof (data->cvs_filestats));
  data->valid_attributes = true;
  
  if ((!S_ISDIR (data->cvs_filestats.st_mode)) && data->is_directory)
  {
    // remove subdir contents - the file type has been changed
    data->contents.clear ();
  }
  
  data->is_directory = S_ISDIR (data->cvs_filestats.st_mode);
}



void cvs_directory_entry::add_entry (cvs_directory_entry &entry)
{
  data->contents.push_back (entry);
}



void cvs_directory_entry::remove_entry (const std::string &name)
{
  directory_list::iterator iter;

  for (iter = data->contents.begin (); iter != data->contents.end (); ++iter)
    if ((*iter).name () == name)
    {
      data->contents.erase (iter);

      return;
    }
}



void cvs_directory_entry::cleanup_tree (time_t t)
{
  directory_list::iterator iter = data->contents.begin ();
  
  while (iter != data->contents.end ())
  {
    if (((*iter).get_scantime () != t) && ((*iter).version () != "0"))
    {
      // we have a discarded entry
      iter = data->contents.erase (iter);	// remove entry and advance to next
    }
    else
    {
      // traverse down the tree
      if ((*iter).is_dir ())
        (*iter).cleanup_tree (t);
	
      ++iter;
    }
  }
}



void cvs_directory_entry::get_directory (directory *dir) const
{
  directory_list::const_iterator iter;
  
  for (iter = data->contents.begin (); iter != data->contents.end (); ++iter)
    dir->push_back ((*iter).name ());
}



bool cvs_directory_entry::get_entry (const std::string &name, cvs_directory_entry &entry) const
{
  directory_list::const_iterator iter;

  for (iter = data->contents.begin (); iter != data->contents.end (); ++iter)
  {
    if ((*iter).name () == name)
    {
      entry = *iter;

      return true;
    }
  }

  return false;
}



bool cvs_directory_entry::get_deep_entry (const std::string &path, cvs_directory_entry &entry) const
{
  std::string request_dir = path;

  // remove leading slash - if any
  if ((request_dir.length () > 0) && (request_dir[0] == '/'))
    request_dir.erase (0, 1);

  if (request_dir.length () == 0)
  {
    entry = *(const_cast<cvs_directory_entry *> (this));

    return true;
  }

  // split path element;
  std::string::size_type pos = request_dir.find ('/');
  if (pos == std::string::npos)
    return get_entry (request_dir, entry);

  std::string remain_dir = request_dir;
  remain_dir.erase (0, pos + 1);
  request_dir.erase (pos);
  
  cvs_directory_entry scan;

  if (!get_entry (request_dir, scan))
    return false;

  return scan.get_deep_entry (remain_dir, entry);
}



bool cvs_directory_entry::get_attributes (struct stat &stbuf) const
{
  if (!data->valid_attributes)
    return false;

  memcpy (&stbuf, &(data->cvs_filestats), sizeof (data->cvs_filestats));

  return true;
}



void cvs_directory_entry::erase_contents ()
{
  data->contents.clear ();
}



void cvs_directory_entry::set_logfilename (const std::string &name)
{
  if (data->log_filename.length () != 0)
  {
    // first delete the old log file
    ::remove (data->log_filename.c_str ());
  }

  data->log_filename = name;
}



cvs_directory_entry::entry_data::entry_data (const std::string & name, const std::string &version, bool isdir)
: refcount (1), cvs_filename (name), cvs_fileversion (version), is_directory (isdir),
  valid_attributes (isdir), last_scanned (0)
{
  memset (&(cvs_filestats), 0, sizeof (cvs_filestats));

  if (isdir)
  {
    // directory entries are always valid
    // because we can not get infos about them from CVS
    cvs_filestats.st_mode = 0777 | S_IFDIR;
  }
}
