/***************************************************************************
 *   Copyright (C) 2004 by Petric Frank                                    *
 *   pfrank@gmx.de                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "cvs-repository.h"
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sstream>
#include <fstream>
#include "cvs-connection.h"
#include "cvs-command-tree.h"
#include "cvs-command-checkout.h"
#include "cvs-command-checkin.h"
#include "cvs-command-add.h"
#include "fs-layer-cvs.h"
#include "fs-layer-checkout.h"
#include "system-log.h"



static char *version_delimiter = "@@";
static char *cvs_directory = "/cvs";
static char *checkout_directory = "/checkout";



cvs_repository::cvs_repository (const std::string &path, cvs_connection *conn,
                                const std::string &repos, const std::string &lab)
: storage_path (path), cvs_storage_path (path + cvs_directory),
  checkout_storage_path (path + checkout_directory), connection (conn),
  repository (repos), label (lab), directory_tree ("/", "", true),
  tree_updater (*conn, directory_tree, lab), refcount (0)
{
}



cvs_repository::~cvs_repository ()
{
  shutdown ();

  delete connection;
}



void cvs_repository::startup ()
{
  std::ifstream is ((storage_path + "/" + label + ".tree").c_str ());
  
  if (is.is_open ())
  {
    is >> directory_tree;
  }

  // start updater thread
  tree_updater.create ();
}



void cvs_repository::shutdown ()
{
  std::ofstream os ((storage_path + "/" + label + ".tree").c_str ());

  if (os.is_open ())
  {
    os << directory_tree;
    os.flush ();
  }
}



fs_layer_cvs *cvs_repository::create_cvs_layer ()
{
  ++refcount;
  
  return new fs_layer_cvs (cvs_storage_path, this);
}



fs_layer_checkout *cvs_repository::create_checkout_layer ()
{
  ++refcount;
  
  return new fs_layer_checkout (checkout_storage_path, this);
}



void cvs_repository::decrement_refcount ()
{
  --refcount;

  if (refcount <= 0)
    delete this;
}



directory *cvs_repository::readdir (const std::string &path)
{
  directory *dir = new directory ();

  // locate the tree element in question
  cvs_directory_entry entry ("", "", false);
  
  if (directory_tree.get_deep_entry (path, entry))
  {
    // load the tree contents
    entry.get_directory (dir);
  }

  return dir;
}



int cvs_repository::getattr (const std::string &path, struct stat *stbuf)
{
  // locate the tree element in question
  cvs_directory_entry entry ("", "", false);

  std::string item = path;
  std::string::size_type pos = item.find (version_delimiter);
  if (pos != std::string::npos)
    item.erase (pos);

  if (!directory_tree.get_deep_entry (path, entry))
  {
    if (pos == std::string::npos)
      return -ENOENT;

    if (!directory_tree.get_deep_entry (item, entry))
      return -ENOENT;

    std::string version = path.substr (pos + strlen (version_delimiter));

    if (entry.version () != version)
    {
      std::string name = path;
      pos = name.rfind ('/');
      if (pos != std::string::npos)
        name.erase (0, pos + 1);

      entry = cvs_directory_entry (name, version, false);
    }
  }
  
  // are the attributes valid ?
  if (!check_entry_attributes (path, entry))
    return -EIO;

  if (!entry.get_attributes (*stbuf))
    return -ENOENT;

  return 0;
}



bool cvs_repository::get_fullname (const std::string &path, std::string *fullpath)
{
  // locate the tree element in question
  cvs_directory_entry entry ("", "", false);

  if (!directory_tree.get_deep_entry (path, entry))
    return false;

  *fullpath = path + version_delimiter + entry.version ();

  return true;
}



int cvs_repository::checkout_file (const std::string &path, char *response, size_t maxsize)
{
  system_log *log = system_log::instance ();
  std::ostringstream errmsg;

  // is there already a checked-out version
  fs_layer *layer = create_checkout_layer ();
  struct stat stbuf;

  if (layer->getattr (path, &stbuf) >= 0)
  {
    delete layer;

    errmsg << "File '" << path << "' already checked out.";
    
    log->error << errmsg.str () << std::endl;

    return handle_response (errmsg.str (), response, maxsize);
  }

  delete layer;

  // locate the tree element in question
  cvs_directory_entry entry ("", "", false);
  
  if (!directory_tree.get_deep_entry (path, entry))
  {
    errmsg << "File '" << path << "' does not exist in CVS view.";

    log->error << errmsg.str () << std::endl;

    return handle_response (errmsg.str (), response, maxsize);
  }

  // is it a file or a directory ?
  if (entry.is_dir ())
  {
    errmsg << "File '" << path << "' is a directory in CVS. CVS does not version directories.";

    log->error << errmsg.str () << std::endl;

    return handle_response (errmsg.str (), response, maxsize);
  }

  // are the attributes valid ?
  if (!check_entry_attributes (path, entry))
  {
    errmsg << "Can not obtain the attributes of file '" << path << "' from CVS.";

    log->error << errmsg.str () << std::endl;

    return handle_response (errmsg.str (), response, maxsize);
  }

  // open the input file
  std::string fullpath = path + version_delimiter + entry.version ();
  std::string inpath = cvs_storage_path + fullpath;
  std::ifstream infile (inpath.c_str ());
  if (!infile.is_open ())
  {
    errmsg << "Can not open file '" << inpath << "' for reading.";

    log->error << errmsg.str () << std::endl;

    return handle_response (errmsg.str (), response, maxsize);
  }

  // create the destination directory - if not existing
  std::string outpath = checkout_storage_path + fullpath;
  std::string outdir = outpath;
  outdir.erase (outdir.rfind ('/'));
  if (!create_directory (outdir, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH))
  {
    errmsg << "Can not create directory '" << outdir << "'.";

    log->error << errmsg.str () << std::endl;

    return handle_response (errmsg.str (), response, maxsize);
  }
  
  // open the output file
  std::ofstream outfile (outpath.c_str ());
  if (!outfile.is_open ())
  {
    errmsg << "Can not open file '" << outpath << "' for writing.";

    log->error << errmsg.str () << std::endl;

    return handle_response (errmsg.str (), response, maxsize);
  }

  // copy the file from cvs to checkout
  char buffer[4096];
  int count;
  
  infile.read (buffer, 4096);
  while ((count = infile.gcount ()) > 0)
  {
    outfile.write (buffer, count);
    infile.read (buffer, 4096);
  }
  
  infile.close ();
  outfile.close ();

  // adjust attributes of new file
  entry.get_attributes (stbuf);
  ::chmod (outpath.c_str (), stbuf.st_mode);

  return 0;
}



int cvs_repository::checkin_file (const std::string &path, char *response, size_t maxsize)
{
  std::ostringstream errmsg;
  system_log *log = system_log::instance ();
  cvs_directory_entry entry;
  std::string inpath = checkout_storage_path;
  std::string shortpath;

  std::string::size_type pos = path.find (version_delimiter);
  if (pos != std::string::npos)
  {
    inpath += path;
    shortpath = path.substr (0, pos);
    std::string version = path.substr (pos + strlen (version_delimiter));

    if (version == "0")
    {
      // file was recently added but not in repository
      struct stat stbuf;

      if (::stat (inpath.c_str (), &stbuf) < 0)
        return -errno;

      std::string::size_type last_slash = path.rfind ('/');
      if (last_slash != std::string::npos)
        entry = cvs_directory_entry (path.substr (last_slash + 1, pos - last_slash - 1), version, S_ISDIR (stbuf.st_mode));
      else
        entry = cvs_directory_entry (path.substr (0, pos), version, S_ISDIR (stbuf.st_mode));
    }
    else
    {
      // locate the tree element in question
      if (!directory_tree.get_deep_entry (path.substr (0, pos), entry))
      {
        errmsg << "File '" << path << "' does not exist in CVS view. Add the file first to CVS.";

        log->error << errmsg.str () << std::endl;

        return handle_response (errmsg.str (), response, maxsize);
      }

      if (version != entry.version ())
      {
        errmsg << "File '" << path << "' was checked out at version " << version
	       << ", but the repository already has version " << entry.version () << ". "
	       << "The local copy must be upgraded.";

        log->error << errmsg.str () << std::endl;

        return handle_response (errmsg.str (), response, maxsize);
      }
    }
  }
  else
  {
    // locate the tree element in question
    if (!directory_tree.get_deep_entry (path, entry))
    {
      errmsg << "File '" << path << "' does not exist in CVS view. Add the file first to CVS.";

      log->error << errmsg.str () << std::endl;

      return handle_response (errmsg.str (), response, maxsize);
    }

    shortpath = path;
    inpath += path + version_delimiter + entry.version ();
  }

  // is it a file or a directory ?
  if (entry.is_dir ())
  {
    errmsg << "File '" << path << "' is a directory in CVS. CVS does not version directories.";

    log->error << errmsg.str () << std::endl;

    return handle_response (errmsg.str (), response, maxsize);
  }

  cvs_command_checkin command (shortpath, inpath, "", entry);

  if (!connection->execute_command (command))
  {
    errmsg << "Checkin of file '" << path << "' failed.";

    log->error << errmsg.str () << std::endl;

    return handle_response (errmsg.str (), response, maxsize);
  }

  // remove checked-out file - the CVS version will pop up
  // maybe we should initiate a re-read of the CVS tree (or at least of this entry)
  ::unlink (inpath.c_str ());

  return 0;
}



int cvs_repository::get_log (const std::string &path, char *response, size_t maxsize)
{
  system_log *log = system_log::instance ();
  std::ostringstream errmsg;
  std::string shortpath;

  std::string::size_type pos = path.find (version_delimiter);
  if (pos != std::string::npos)
  {
    if (path.substr (pos + strlen (version_delimiter)) == "0")
      return 0;		// do not return log for recently added items

    shortpath = path.substr (0, pos);
  }
  else
    shortpath = path;

  // locate the tree element in question
  cvs_directory_entry entry;
  
  if (!directory_tree.get_deep_entry (shortpath, entry))
  {
    errmsg << "'" << path << "' is not a CVS item." << std::endl;

    log->error << errmsg.str () << std::endl;

    return handle_response (errmsg.str (), response, maxsize);
  }

  std::string fname = entry.get_logfilename ();

  if (fname.length () == 0)
  {
    errmsg << "No log file name assigned to '" << path << "'" << std::endl;

    log->error << errmsg.str () << std::endl;

    return handle_response (errmsg.str (), response, maxsize);
  }

  struct stat stbuf;
  
  if (::stat (fname.c_str (), &stbuf) < 0)
    return -errno;

  if (maxsize == 0)
    return stbuf.st_size;

  if (maxsize < stbuf.st_size)
    return -ERANGE;

  // copy file contents to buffer
  int count;
  char *ptr = response;
  std::ifstream in (fname.c_str ());
  if (!in.is_open ())
  {
    errmsg << "Can not open log file for reading" << std::endl;

    log->error << errmsg.str () << std::endl;

    return handle_response (errmsg.str (), response, maxsize);
  }

  in.read (ptr, 4096);
  while ((count = in.gcount ()) > 0)
  {
    ptr += count;
    in.read (ptr, 4096);
  }
  
  in.close ();

  return stbuf.st_size;
}



int cvs_repository::add_item (const std::string &path, char *response,
                              size_t maxsize, fs_layer *private_layer)
{
  std::ostringstream errmsg;
  system_log *log = system_log::instance ();

  // split directory and name
  std::string dir;
  std::string name = path;
  std::string::size_type pos = name.rfind ('/');
  if (pos != std::string::npos)
  {
    dir = name;
    dir.erase (pos);
    name.erase (0, pos + 1);
  }

  // get insert location - also check wheter parent directory is part of CVS
  cvs_directory_entry insert_location ("", "", false);
  if (!directory_tree.get_deep_entry (dir, insert_location))
  {
    errmsg << "Parent directory of entry of '" << dir << "' not in cache tree.";

    log->error << errmsg.str () << std::endl;

    return handle_response (errmsg.str (), response, maxsize);
  }

  // obtain the attributes of the file/directory to add
  struct stat stbuf;
  if (private_layer->getattr (path, &stbuf) < 0)
  {
    errmsg << "Can not obtain file info for file '" << path << "'.";

    log->error << errmsg.str () << std::endl;

    return handle_response (errmsg.str (), response, maxsize);
  }

//  cvs_directory_entry new_entry ("", "", false);

  if (S_ISDIR (stbuf.st_mode))
  {
    // create the directory in checkout layer
    if (!create_directory (checkout_storage_path + path, S_IRUSR | S_IWUSR))
    {
      errmsg << "Create of directory '" << path << "' failed.";

      log->error << errmsg.str () << std::endl;

      return handle_response (errmsg.str (), response, maxsize);
    }
    
    cvs_directory_entry new_entry = cvs_directory_entry (name, "", true);

    // add the directory to CVS - they are immediately added to the repository
    cvs_command_add command (path, "", new_entry);
    if (!connection->execute_command (command))
    {
      errmsg << "Add of directory '" << path << "' to CVS failed.";

      log->error << errmsg.str () << std::endl;

      ::rmdir ((checkout_storage_path + path).c_str ());

      return handle_response (errmsg.str (), response, maxsize);
    }

    // add entry to the cache tree
    insert_location.add_entry (new_entry);
  }
  else
  {
    // create the parent directory in checkout layer
    if (!create_directory (checkout_storage_path + dir, S_IRUSR | S_IWUSR))
    {
      errmsg << "Create of parent directory '" << path << "' failed.";

      log->error << errmsg.str () << std::endl;

      return handle_response (errmsg.str (), response, maxsize);
    }

    // move the file to the new location
    if (::rename ((private_layer->get_basedir () + path).c_str (), 
                  (checkout_storage_path + dir + "/" + name + version_delimiter + "0").c_str ()) < 0)
      return -errno;
  }

  return 0;
}



bool cvs_repository::check_entry_attributes (const std::string &path, cvs_directory_entry &entry)
{
  if (!entry.valid ())
  {
    std::string version = entry.version ();
    std::string shortpath = path;
    std::string::size_type pos = shortpath.find (version_delimiter);
    if (pos != std::string::npos)
    {
      version = path;
      version.erase (0, pos + strlen (version_delimiter));
      shortpath.erase (pos);
    }

    cvs_command_checkout command (shortpath, cvs_storage_path + shortpath + version_delimiter + version, entry);

    return connection->execute_command (command);
  }
  
  return true;
}



bool cvs_repository::create_directory (const std::string &dir, mode_t mode)
{
  std::string path = dir;
  std::string::size_type pos;
  std::string dirpart;
  int dirmode = S_IFDIR | mode;

  if ((mode & S_IRUSR) != 0)
    dirmode |= S_IXUSR;

  if ((mode & S_IRGRP) != 0)
    dirmode |= S_IXGRP;

  if ((mode & S_IROTH) != 0)
    dirmode |= S_IXOTH;

  pos = 0;
  do
  {
    pos = path.find ('/', pos + 1);

    dirpart = path;
    if (pos != std::string::npos)
      dirpart.erase (pos);

    struct stat info;

    if (::lstat (dirpart.c_str (), &info) != 0)
    {
      int result = ::mkdir (dirpart.c_str (), dirmode);
      if (result != 0)
        return false;
    }
    else
      if (!S_ISDIR (info.st_mode))
        return false;

  } while (pos != std::string::npos);

  return true;
}



int cvs_repository::handle_response (const std::string &response, char *buffer, size_t maxsize)
{
  if (maxsize != 0)
  {
    if (maxsize < response.length ())
      return -ERANGE;

    strcpy (buffer, response.c_str ());
  }

  return response.length () + 1;
}
