/***************************************************************************
 *   Copyright (C) 2004 by Petric Frank                                    *
 *   pfrank@gmx.de                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "fuse-base.h"
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <iostream>
#include "fuse-callback.h"



fuse_base::fuse_base ()
{
  callback = fuse_callback::instance ();
}



fuse_base::~fuse_base ()
{
}



void fuse_base::usage(const std::string &progname)
{
  help_cmdline (progname);

  std::cerr << "Options:" << std::endl;
 
  help_options ();

  std::cerr << std::endl
            << "Mount options:" << std::endl;

  help_mount_options ();
}



void fuse_base::help_cmdline (const std::string &progname)
{
  std::cerr << "usage: " << progname << " mountpoint [options]" << std::endl;
}



void fuse_base::help_options ()
{
  std::cerr << "    -d                  enable debug output (implies -f)" << std::endl
            << "    -f                  foreground operation" << std::endl
            << "    -s                  disable multithreaded operation" << std::endl
            << "    -o opt,[opt...]     mount options" << std::endl
            << "    -h                  print help" << std::endl;
}



void fuse_base::help_mount_options ()
{
  std::cerr << "    default_permissions    enable permission checking" << std::endl
            << "    allow_other            allow access to other users" << std::endl
            << "    kernel_cache           cache files in kernel" << std::endl
            << "    large_read             issue large read requests (2.4 only)" << std::endl
            << "    direct_io              use direct I/O" << std::endl
            << "    max_read=N             set maximum size of read requests" << std::endl
            << "    hard_remove            immediate removal (don't hide files)" << std::endl
            << "    debug                  enable debug output" << std::endl
            << "    fsname=NAME            set filesystem name in mtab" << std::endl;
}



void fuse_base::run (int argc, char *argv[])
{
  for (int loop = 1; loop < argc; ++loop)
    if (strcmp (argv[loop], "-h") == 0)
    {
      usage (argv[0]);

      return;
    }

  if (!run_prepare (argc, argv))
    return;

  callback->setbase (this);

  fuse_main (argc, argv, &fuse_callback::oper);
}



bool fuse_base::run_prepare (int &argc, char *argv[])
{
  return true;
}



// *********************************************************************
// The following implementations are dummy implementations inspired by
// 'hello.c' of the fuse-1.9 package.
// Each derivation from this class should implement the methods
// themself.
// *********************************************************************

int fuse_base::getattr (const char *path, struct stat *stbuf)
{
  memset (stbuf, 0, sizeof (struct stat));
  
  if (strcmp (path, "/") == 0)
  {
    // root directory
    stbuf->st_mode = S_IFDIR | 0755;
    stbuf->st_nlink = 2;
    
    return 0;
  }

  if (strcmp (path, "/hello") == 0)
  {
    stbuf->st_mode = S_IFREG | 0444;
    stbuf->st_nlink = 1;
    stbuf->st_size = 0;
    stbuf->st_uid = 1000;
    
    return 0;
  }
  
  return -ENOENT;
}



int fuse_base::readlink (const char *path, char *buf, size_t size)
{
  return -ENOENT;
}



int fuse_base::getdir (const char *path, fuse_dirh_t h, fuse_dirfil_t filler)
{
  if (strcmp (path, "/") != 0)
  {
    // not root directory
    return -ENOENT;
  }

  filler (h, ".", 0);
  filler (h, "..", 0);
  filler (h, "hello", 0);

  return 0;
}



int fuse_base::mknod (const char *path, mode_t mode, dev_t rdev)
{
  return -ENOENT;
}



int fuse_base::mkdir (const char *path, mode_t mode)
{
  return -ENOENT;
}



int fuse_base::unlink (const char *path)
{
  return -ENOENT;
}



int fuse_base::rmdir (const char *path)
{
  return -ENOENT;
}



int fuse_base::symlink (const char *path, const char *to)
{
  return -ENOENT;
}



int fuse_base::rename (const char *path, const char *to)
{
  return -ENOENT;
}



int fuse_base::link (const char *path, const char *to)
{
  return -ENOENT;
}



int fuse_base::chmod (const char *path, mode_t mode)
{
  return -ENOENT;
}



int fuse_base::chown (const char *path, uid_t uid, gid_t gid)
{
  return -ENOENT;
}



int fuse_base::truncate (const char *path, off_t size)
{
  return -ENOENT;
}



int fuse_base::utime (const char *path, struct utimbuf *buf)
{
  return -ENOENT;
}



int fuse_base::open (const char *path, int flags)
{
  if (strcmp (path, "/hello") != 0)
  {
    // not the dummy file
    return -ENOENT;
  }

  if ((flags & 3) != O_RDONLY)
  {
    // not read-only access
    return -EACCES;
  }

  return 0;
}



int fuse_base::read (const char *path, char *buf, size_t size, off_t offset)
{
  if (strcmp (path, "/hello") != 0)
  {
    // not the dummy file
    return -ENOENT;
  }

  return 0;
}



int fuse_base::write (const char *path, const char *buf, size_t size, off_t offset)
{
  return -ENOENT;
}



int fuse_base::statfs (const char *path, struct statfs *stbuf)
{
  return -ENOENT;
}



int fuse_base::flush (const char *path)
{
  return -ENOENT;
}



int fuse_base::release (const char *path, int mode)
{
  return -ENOENT;
}



int fuse_base::fsync (const char *path, int mode)
{
  return -ENOENT;
}



int fuse_base::setxattr (const char *path, const char *name, const char *value, size_t size, int flags)
{
  return -ENOENT;
}



int fuse_base::getxattr (const char *path, const char *name, char *value, size_t size)
{
  return -ENOENT;
}



int fuse_base::listxattr (const char *path, char *list, size_t size)
{
  return -ENOENT;
}



int fuse_base::removexattr (const char *path, const char *name)
{
  return -ENOENT;
}
